% This function sets up some basic BIEs and verify the accuracy of
% the resulting solver.
%
% All routines concern the basic double layer potential for Laplace.
%
% Nystrom discretization with Composite Gaussian quadrature is used.
%
% We use "Bremer" style weighting of the matrix.

function main01

rng('default')
rng(0)

fprintf(1,'Executing a test for an exterior problems for a single CE.\n')

%%% Construct the contour.
Npan      = 20;   % Number of panels along the contour.
Ngau      = 10;   % Number of points per panel.
flag_ext  = 1;
[CE,xx_int,xx_ext] = LOCAL_geometry_elli(Npan,Ngau,flag_ext);
ww   = CE(10,:);
ntot = size(CE,2);
fprintf(1,'Total number of nodes (ntot) = %d\n',ntot)

%%% Construct the system matrix.
A = LOCAL_get_A_diag(CE,1:ntot);

%%% Test the functions for computing submatrices.
[~,ind] = sort(rand(1,ntot));
ind1    = sort(ind(1:5));
ind2    = LOCAL_index_complement(ind1,ntot);
A11     = LOCAL_get_A_diag(CE,ind1);
fprintf(1,'Max discrepancy in A11 = %12.5e\n',...
        max(max(abs(A11 - A(ind1,ind1)))))
A12     = LOCAL_get_A_offd(CE,ind1,ind2);
fprintf(1,'Max discrepancy in A12 = %12.5e\n',...
        max(max(abs(A12 - A(ind1,ind2)))))

%%% Set up and solve a reference problem.
qq_int    = randn(size(xx_int,2),1);
u_dir     = LOCAL_pot_from_charges(CE([1,4],:),xx_int,qq_int);
u_ref     = LOCAL_pot_from_charges(xx_ext,xx_int,qq_int);
sww       = sqrt(ww);
sigma     = (1./sww').*(A\(sww'.*u_dir));
u_app     = LOCAL_eval_pot(xx_ext,CE,sigma);
fprintf(1,'   ||u_ref - u_app|| / ||u_app|| = %18.10e\n',...
        norm(u_ref - u_app)/norm(u_app))

%%% Plot the geometry.
figure(1)
plot(CE(1,:),CE(4,:),'.',...
     xx_int(1,:),xx_int(2,:),'rx',...
     xx_ext(1,:),xx_ext(2,:),'bx')
legend('C','xxint','xxext')
axis equal

keyboard

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [CE,xx_int,xx_ext] = LOCAL_geometry_elli(Npan,Ngau,flag_ext)

r1            = 2;    % Parameter controlling the geometry.
r2            = 0.5;  % Parameter controlling the geometry. 
t_sta         = 0;    % Start point of the discretization.
t_end         = 2*pi; % End   point of the discretization. 
tt_panels     = linspace(t_sta,t_end,Npan+1);
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);
C             = zeros(6,length(tt));
C(1,:)        =  r1*cos(tt);
C(2,:)        = -r1*sin(tt);
C(3,:)        = -r1*cos(tt);
C(4,:)        =  r2*sin(tt);
C(5,:)        =  r2*cos(tt);
C(6,:)        = -r2*sin(tt);
ww            = ww_gauss.*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.5*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.5*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

ntot      = size(C,2);
if (flag_ext == 1)
  zz        = [0;0];
  CE        = [C;...
               zz(1)*ones(1,ntot);...
               zz(2)*ones(1,ntot);...
               -ones(1,ntot);...
               ww];
else
  zz = 1.5*R_max*[1/sqrt(2);1/sqrt(2)];
  CE        = [C;...
               zz(1)*ones(1,ntot);...
               zz(2)*ones(1,ntot);...
               ones(1,ntot);...
               ww];
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [tt,ww] = LOCAL_get_gauss_nodes(tt_panels,Ngau)

npanels = length(tt_panels)-1;
tt   = zeros(1,npanels*Ngau);
ww   = zeros(1,npanels*Ngau);
[t_ref,w_ref] = LOCAL_lgwt(Ngau,0,1);
t_ref         = t_ref(end:(-1):1);
w_ref         = w_ref(end:(-1):1);
for i = 1:npanels
  h       = tt_panels(i+1) - tt_panels(i);
  ind     = (i-1)*Ngau + (1:Ngau);
  tt(ind) = tt_panels(i) + h*t_ref;
  ww(ind) = h*w_ref;
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [x,w] = LOCAL_lgwt(N,a,b)

%%% Use tabulated values if they exist
if (N == 5)
  TMP = [0.000000000000000,  0.568888888888889;...
         0.538469310105683,  0.478628670499366;...
         0.906179845938664,  0.236926885056189];
  xx = [TMP(end:(-1):2,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):2,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(2:end,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(2:end,2)'];
elseif (N == 10)
  TMP = [0.148874338981631211,  0.295524224714752870;...
         0.433395394129247191,  0.269266719309996355;...
         0.679409568299024406,  0.219086362515982044;...
         0.865063366688984511,  0.149451349150580593;...
         0.973906528517171720,  0.066671344308688138];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
elseif (N == 20)
  TMP = [0.076526521133497333755,  0.152753387130725850698; ...
         0.227785851141645078080,  0.149172986472603746788; ...
         0.373706088715419560673,  0.142096109318382051329; ...
         0.510867001950827098004,  0.131688638449176626898; ...
         0.636053680726515025453,  0.118194531961518417312; ...
         0.746331906460150792614,  0.101930119817240435037; ...
         0.839116971822218823395,  0.083276741576704748725; ...
         0.912234428251325905868,  0.062672048334109063570; ...
         0.963971927277913791268,  0.040601429800386941331; ...
         0.993128599185094924786,  0.017614007139152118312];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
end
if exist('xx','var')
  xmid  = 0.5*(a+b);
  scale = 0.5*(b-a);
  x     = xmid + scale*xx';
  w     =        scale*ww';
  return
end

% lgwt.m
%
% This script is for computiNgau definite integrals usiNgau Legendre-Gauss 
% Quadrature. Computes the Legendre-Gauss nodes and weights  on an interval
% [a,b] with truncation order N
%
% Suppose you have a continuous function f(x) which is defined on [a,b]
% which you can evaluate at any x in [a,b]. Simply evaluate it at all of
% the values contained in the x vector to obtain a vector f. Then compute
% the definite integral usiNgau sum(f.*w);
%
% Written by Greg von Winckel - 02/25/2004
N=N-1;
N1=N+1; N2=N+2;

xu=linspace(-1,1,N1)';

% Initial guess
y=cos((2*(0:N)'+1)*pi/(2*N+2))+(0.27/N1)*sin(pi*xu*N/N2);

% Legendre-Gauss Vandermonde Matrix
L=zeros(N1,N2);

% Derivative of LGVM
Lp=zeros(N1,N2);

% Compute the zeros of the N+1 Legendre Polynomial
% usiNgau the recursion relation and the Newton-Raphson method

y0=2;

% Iterate until new points are uniformly within epsilon of old points
while max(abs(y-y0))>eps
    
    
    L(:,1)=1;
    Lp(:,1)=0;
    
    L(:,2)=y;
    Lp(:,2)=1;
    
    for k=2:N1
        L(:,k+1)=( (2*k-1)*y.*L(:,k)-(k-1)*L(:,k-1) )/k;
    end
 
    Lp=(N2)*( L(:,N1)-y.*L(:,N2) )./(1-y.^2);   
    
    y0=y;
    y=y0-L(:,N2)./Lp;
    
end

% Linear map from[-1,1] to [a,b]
x=(a*(1-y)+b*(1+y))/2;      

% Compute the weights
w=(b-a)./((1-y.^2).*Lp.^2)*(N2/N1)^2;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_pot_from_charges(xx_trg,xx_src,qq)

ntrg  = size(xx_trg,2);
nsrc  = size(xx_src,2);
dist1 = (xx_trg(1,:)' * ones(1,nsrc)) - ones(ntrg,1)*xx_src(1,:);
dist2 = (xx_trg(2,:)' * ones(1,nsrc)) - ones(ntrg,1)*xx_src(2,:);

distsq = dist1.*dist1 + dist2.*dist2;

u = -(1/(4*pi))*log(distsq)*reshape(qq,nsrc,1);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A_diag(CE,ind)

beta = -1/(2*pi);
n    = length(ind);
ww   = CE(10,ind);
dd1  = CE(1,ind)' * ones(1,n) - ones(n,1) * CE(1,ind);
dd2  = CE(4,ind)' * ones(1,n) - ones(n,1) * CE(4,ind);
nn1  = ones(n,1)*( CE(5,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
nn2  = ones(n,1)*(-CE(2,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
ddsq  = (dd1.*dd1 + dd2.*dd2) + eye(n);

% Compute the "pure" double layer kernel (note that diagonal entries will be zero).
D     = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Add the diagonal term.
d_vec = -(-CE(6,ind).*CE(2,ind) + CE(3,ind).*CE(5,ind))./((4*pi)*(CE(2,ind).^2 + CE(5,ind).^2).^1.5);
D     = D + diag(d_vec);

% Add the contribution from the "correction" sources.
ddsq  = (CE(1,ind)' * ones(1,n) - ones(n,1)*CE(7,ind)).^2 + ...
        (CE(4,ind)' * ones(1,n) - ones(n,1)*CE(8,ind)).^2;
D     = D + 0.5*beta*log(ddsq); 

% Scale by the root of the quadrature weights.
A = diag(0.5*CE(9,ind)) + sqrt(ww' * ww).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A_offd(CE,ind1,ind2)

beta = -1/(2*pi);
n1   = length(ind1);
n2   = length(ind2);
dd1  = CE(1,ind1)' * ones(1,n2) - ones(n1,1) * CE(1,ind2);
dd2  = CE(4,ind1)' * ones(1,n2) - ones(n1,1) * CE(4,ind2);
nn1  = ones(n1,1)*( CE(5,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
nn2  = ones(n1,1)*(-CE(2,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
ddsq = dd1.*dd1 + dd2.*dd2;

% Compute the "pure" double layer kernel.
D    = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Add the contribution from the "correction" sources.
ddsq = (CE(1,ind1)' * ones(1,n2) - ones(n1,1)*CE(7,ind2)).^2 + ...
       (CE(4,ind1)' * ones(1,n2) - ones(n1,1)*CE(8,ind2)).^2;
D    = D + 0.5*beta*log(ddsq);
     
% Scale by the root of the quadrature weights.
ww1  = CE(10,ind1);
ww2  = CE(10,ind2);
A    = (sqrt(ww1'*ww2)).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_eval_pot(xx_trg,CE,sigma)

C = CE(1:6,:);
ww = CE(10,:);
beta = -1/(2*pi);

n_trg = size(xx_trg,2);
n_src = size(C,2);

dd1 = xx_trg(1,:)' * ones(1,n_src) - ones(n_trg,1) * C(1,:);
dd2 = xx_trg(2,:)' * ones(1,n_src) - ones(n_trg,1) * C(4,:);
nn1 = ones(n_trg,1)*( C(5,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));
nn2 = ones(n_trg,1)*(-C(2,:)./sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:)));

D = -(1/(2*pi))*(nn1.*dd1 + nn2.*dd2)./(dd1.*dd1 + dd2.*dd2);

ddsq = (xx_trg(1,:)' * ones(1,n_src) - ones(n_trg,1) * CE(7,:)).^2 + ...
       (xx_trg(2,:)' * ones(1,n_src) - ones(n_trg,1) * CE(8,:)).^2;

u = (D + 0.5*beta*log(ddsq)) * (ww' .* sigma);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ind2    = LOCAL_index_complement(ind1,ntot)

ind       = 1:ntot;
ind(ind1) = 2*ntot*ones(1,length(ind1));
ind       = sort(ind);
ind2      = ind(1:(ntot-length(ind1)));

return
