% This function sets up a basic BIEs on the "eye" contour.
% It then computes basis matrices using the "Green" trick
% and verifies the accuracy of the basis matrices.
%
% All routines concern the basic double layer potential for Laplace.
%
% Nystrom discretization with Composite Gaussian quadrature is used.
%
% We use "Bremer" style weighting of the matrix.

function main21()

%%% Set discretization parameters.
Npan = 8;
Ngau = 10;
Nref = 3;
acc  = 1e-10;
flag_ext  = 1;

%%% Construct the contour.
[CE,xx_int,xx_ext] = LOCAL_geometry_eye(Npan,Ngau,Nref,flag_ext);

%%% Partition the contour into three sets.
h  = 2/Npan;
I1 = find(CE(1,:) < (-1+h));
I2 = find((CE(1,:)>(-1+h)) & (CE(1,:)<(1-h)));
I3 = find(CE(1,:) > ( 1-h));

%%% Plot the contour and the three sets.
plot(CE(1,[1:end,1]),CE(4,[1:end,1]),'k-',...
     CE(1,I2),CE(4,I2),'g.',...
     CE(1,I1),CE(4,I1),'r.',...
     CE(1,I3),CE(4,I3),'b.')
axis equal
axis off
%print -depsc Pics/fig_eye.eps

%%% Construct a skeleton for I1.
[T,J] = LOCAL_get_bases_green(CE,I1,acc);
k     = size(T,1);

%%% Validate the new skeletons
I1comp = LOCAL_index_complement(I1,size(CE,2));
A21 = LOCAL_get_A_offd(CE,I1comp,I1);
A12 = LOCAL_get_A_offd(CE,I1,I1comp);
Js  = J(1:k);
Jr  = J((k+1):end);
fprintf(1,'Error in skeleton for A21 = %12.5e\n',max(max(abs(A21(:,Jr) - A21(:,Js)* T))))
fprintf(1,'Error in skeleton for A12 = %12.5e\n',max(max(abs(A12(Jr,:) - T'*A12(Js,:)))))

%%% Compare the rank from the Green trick to the optimal rank.
[TT,JJ] = id_decomp([A21;A12'],acc);
fprintf(1,'The optimal rank of interaction = %d\n',size(TT,1))
fprintf(1,'The rank from the Green trick   = %d\n',size(T ,1))

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [CE,xx_int,xx_ext] = LOCAL_geometry_eye(Npan,Ngau,Nref,flag_ext)

if (mod(Npan,2) == 1)
  fprintf(1,'ERROR: Npan must be even.\n')
  keyboard
end

h             = 2/Npan;
tt_panels     = [-1,-1 + h*(1./(2.^(Nref:(-1):1))),...
                 linspace(-1+h,1-h,Npan-1),...
                  1 - h*(1./(2.^(1:Nref))),1];
[tt,ww_gauss] = LOCAL_get_gauss_nodes(tt_panels,Ngau);

H             = 0.25;
nhalf          = length(tt);
CU             = zeros(6,nhalf);
CU(1,:)        = tt;
CU(2,:)        = ones(size(tt));
CU(3,:)        = zeros(size(tt));
CU(4,:)        = H*(1 - tt.*tt);
CU(5,:)        = -2*H*tt;
CU(6,:)        = -2*H*ones(size(tt));
CB             = zeros(6,nhalf);
CB(1,:)        = -tt;
CB(2,:)        = -ones(size(tt));
CB(3,:)        = zeros(size(tt));
CB(4,:)        = -H*(1 - tt.*tt);
CB(5,:)        = 2*H*tt;
CB(6,:)        = 2*H*ones(size(tt));
C              = [CU,CB];
ww            = [ww_gauss,ww_gauss].*sqrt(C(2,:).*C(2,:) + C(5,:).*C(5,:));
ntot          = 2*nhalf;
CE            = [C;...
                 zeros(2,ntot);...
                 -ones(1,ntot);...
                 ww];

nloc   = 10;
R_min  = sqrt(min(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
R_max  = sqrt(max(C(1,:).*C(1,:) + C(4,:).*C(4,:)));
xx_int = [0.4*R_min*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          0.4*R_min*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];
xx_ext = [1.3*R_max*cos(linspace(0,2*pi*(1 - 1/nloc),nloc));...
          1.3*R_max*sin(linspace(0,2*pi*(1 - 1/nloc),nloc))];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [tt,ww] = LOCAL_get_gauss_nodes(tt_panels,Ngau)

npanels = length(tt_panels)-1;
tt   = zeros(1,npanels*Ngau);
ww   = zeros(1,npanels*Ngau);
[t_ref,w_ref] = LOCAL_lgwt(Ngau,0,1);
t_ref         = t_ref(end:(-1):1);
w_ref         = w_ref(end:(-1):1);
for i = 1:npanels
  h       = tt_panels(i+1) - tt_panels(i);
  ind     = (i-1)*Ngau + (1:Ngau);
  tt(ind) = tt_panels(i) + h*t_ref;
  ww(ind) = h*w_ref;
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [x,w] = LOCAL_lgwt(N,a,b)

%%% Use tabulated values if they exist
if (N == 5)
  TMP = [0.000000000000000,  0.568888888888889;...
         0.538469310105683,  0.478628670499366;...
         0.906179845938664,  0.236926885056189];
  xx = [TMP(end:(-1):2,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):2,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(2:end,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(2:end,2)'];
elseif (N == 10)
  TMP = [0.148874338981631211,  0.295524224714752870;...
         0.433395394129247191,  0.269266719309996355;...
         0.679409568299024406,  0.219086362515982044;...
         0.865063366688984511,  0.149451349150580593;...
         0.973906528517171720,  0.066671344308688138];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
elseif (N == 20)
  TMP = [0.076526521133497333755,  0.152753387130725850698; ...
         0.227785851141645078080,  0.149172986472603746788; ...
         0.373706088715419560673,  0.142096109318382051329; ...
         0.510867001950827098004,  0.131688638449176626898; ...
         0.636053680726515025453,  0.118194531961518417312; ...
         0.746331906460150792614,  0.101930119817240435037; ...
         0.839116971822218823395,  0.083276741576704748725; ...
         0.912234428251325905868,  0.062672048334109063570; ...
         0.963971927277913791268,  0.040601429800386941331; ...
         0.993128599185094924786,  0.017614007139152118312];
  xx = [TMP(end:(-1):1,1)',-TMP(:,1)'];
  ww = [TMP(end:(-1):1,2)', TMP(:,2)'];
%  xx = [-TMP(end:(-1):1,1)',TMP(:,1)'];
%  ww = [ TMP(end:(-1):1,2)',TMP(:,2)'];
end
if exist('xx','var')
  xmid  = 0.5*(a+b);
  scale = 0.5*(b-a);
  x     = xmid + scale*xx';
  w     =        scale*ww';
  return
end

% lgwt.m
%
% This script is for computiNgau definite integrals usiNgau Legendre-Gauss 
% Quadrature. Computes the Legendre-Gauss nodes and weights  on an interval
% [a,b] with truncation order N
%
% Suppose you have a continuous function f(x) which is defined on [a,b]
% which you can evaluate at any x in [a,b]. Simply evaluate it at all of
% the values contained in the x vector to obtain a vector f. Then compute
% the definite integral usiNgau sum(f.*w);
%
% Written by Greg von Winckel - 02/25/2004
N=N-1;
N1=N+1; N2=N+2;

xu=linspace(-1,1,N1)';

% Initial guess
y=cos((2*(0:N)'+1)*pi/(2*N+2))+(0.27/N1)*sin(pi*xu*N/N2);

% Legendre-Gauss Vandermonde Matrix
L=zeros(N1,N2);

% Derivative of LGVM
Lp=zeros(N1,N2);

% Compute the zeros of the N+1 Legendre Polynomial
% usiNgau the recursion relation and the Newton-Raphson method

y0=2;

% Iterate until new points are uniformly within epsilon of old points
while max(abs(y-y0))>eps
    
    
    L(:,1)=1;
    Lp(:,1)=0;
    
    L(:,2)=y;
    Lp(:,2)=1;
    
    for k=2:N1
        L(:,k+1)=( (2*k-1)*y.*L(:,k)-(k-1)*L(:,k-1) )/k;
    end
 
    Lp=(N2)*( L(:,N1)-y.*L(:,N2) )./(1-y.^2);   
    
    y0=y;
    y=y0-L(:,N2)./Lp;
    
end

% Linear map from[-1,1] to [a,b]
x=(a*(1-y)+b*(1+y))/2;      

% Compute the weights
w=(b-a)./((1-y.^2).*Lp.^2)*(N2/N1)^2;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [T,J] = LOCAL_get_bases_green(CE,ind,acc)

%%% Set various parameters that balance the accuracy and efficiency.
proxyscale = 3;       % The relative size of the "Green box".
nside      = 25;      % The number of nodes on one side of the proxy box.

%%% Construct a box that encloses I1.
x1min = min(CE(1,ind));
x1max = max(CE(1,ind));
x2min = min(CE(4,ind));
x2max = max(CE(4,ind));
x1c   = 0.5*(x1min+x1max);
x2c   = 0.5*(x2min+x2max);
a     = 0.5*max([x1max-x1min,x2max-x2min]);

%%% Find the near-field
indcomp = LOCAL_index_complement(ind,size(CE,2));
J       = find( (CE(1,indcomp) < (x1c+proxyscale*a)) & ...
                (CE(1,indcomp) > (x1c-proxyscale*a)) & ...
                (CE(4,indcomp) < (x2c+proxyscale*a)) & ...
                (CE(4,indcomp) > (x2c-proxyscale*a)) );
indnear   = indcomp(J);          

%%% Construct the proxy box.
A = a*proxyscale;
h = 2*A/nside;
xxproxy_e = [x1c + A*ones(1,nside);       x2c + linspace(-A,A-h,nside)];
xxproxy_n = [x1c + linspace(A,-A+h,nside);x2c + A*ones(1,nside)];
xxproxy_w = [x1c - A*ones(1,nside);       x2c + linspace(A,-A+h,nside)];
xxproxy_s = [x1c + linspace(-A,A-h,nside);x2c - A*ones(1,nside)];
xxproxy   = [xxproxy_e,xxproxy_n,xxproxy_w,xxproxy_s];

%%% Construct the proxy fields.
A_proxy_ind   = LOCAL_A_to_proxy(CE,xxproxy,ind);
A_ind_proxy = LOCAL_A_from_proxy(CE,ind,xxproxy);

%%% Construct the contributions from the near field.
A_near_ind = LOCAL_get_A_offd(CE,indnear,ind);
A_ind_near = LOCAL_get_A_offd(CE,ind,indnear);

%%% Assemble all interactions and compress.
X = [A_near_ind;...
     A_ind_near';...
     A_proxy_ind;...
     A_ind_proxy'];
[T,J] = id_decomp(X,acc);

%%% Plot to illustrate the geometry of the local compression.
figure(1)
plot(CE(1,:),CE(4,:),'k.',...
     CE(1,ind),CE(4,ind),'r.',...
     CE(1,indnear),CE(4,indnear),'b.',...
     xxproxy(1,:),xxproxy(2,:),'m.',...
     x1c + a*[-1,1,1,-1,-1],x2c + a*[-1,-1,1,1,-1],'g')
axis equal
legend('Full contour','Target patch','Near field','Proxy','Enclosing box')

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ind2 = LOCAL_index_complement(ind1,ntot)

ind       = 1:ntot;
ind(ind1) = 2*ntot*ones(1,length(ind1));
ind       = sort(ind);
ind2      = ind(1:(ntot-length(ind1)));

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_A_from_proxy(CE,ind,xxproxy)

beta = -1/(2*pi);
n1   = length(ind);
n2   = size(xxproxy,2);
dd1  = CE(1,ind)' * ones(1,n2) - ones(n1,1) * xxproxy(1,:);
dd2  = CE(4,ind)' * ones(1,n2) - ones(n1,1) * xxproxy(2,:);
ddsq = dd1.*dd1 + dd2.*dd2;

% Compute a single layer kernel.
D    = beta*0.5*log(ddsq);

% Scale by the root of the quadrature weights.
ww  = CE(10,ind);
A   = (sqrt(ww')*ones(1,n2)).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_A_to_proxy(CE,xxproxy,ind)

beta = -1/(2*pi);
n1   = size(xxproxy,2);
n2   = length(ind);
dd1  = xxproxy(1,:)' * ones(1,n2) - ones(n1,1) * CE(1,ind);
dd2  = xxproxy(2,:)' * ones(1,n2) - ones(n1,1) * CE(4,ind);
nn1  = ones(n1,1)*( CE(5,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
nn2  = ones(n1,1)*(-CE(2,ind)./sqrt(CE(2,ind).*CE(2,ind) + CE(5,ind).*CE(5,ind)));
ddsq = dd1.*dd1 + dd2.*dd2;

% Compute the unadultered double layer kernel.
D    = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Scale by the root of the quadrature weights.
ww  = CE(10,ind);
A   = D.*(ones(n1,1)*sqrt(ww));

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A_offd(CE,ind1,ind2)

beta = -1/(2*pi);
n1   = length(ind1);
n2   = length(ind2);
dd1  = CE(1,ind1)' * ones(1,n2) - ones(n1,1) * CE(1,ind2);
dd2  = CE(4,ind1)' * ones(1,n2) - ones(n1,1) * CE(4,ind2);
nn1  = ones(n1,1)*( CE(5,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
nn2  = ones(n1,1)*(-CE(2,ind2)./sqrt(CE(2,ind2).*CE(2,ind2) + CE(5,ind2).*CE(5,ind2)));
ddsq = dd1.*dd1 + dd2.*dd2;

% Compute the "pure" double layer kernel.
D    = beta*(nn1.*dd1 + nn2.*dd2)./ddsq;

% Add the contribution from the "correction" sources.
ddsq = (CE(1,ind1)' * ones(1,n2) - ones(n1,1)*CE(7,ind2)).^2 + ...
       (CE(4,ind1)' * ones(1,n2) - ones(n1,1)*CE(8,ind2)).^2;
D    = D + 0.5*beta*log(ddsq);
     
% Scale by the root of the quadrature weights.
ww1  = CE(10,ind1);
ww2  = CE(10,ind2);
A    = (sqrt(ww1'*ww2)).*D;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

