% This function tests the HBS compression scheme on a torus-like domain in 3D.
% (It is not actually a torus, and is not rotationally symmetric.)
%
% The mesh is subdivided into a quad-tree in PARAMETER space.
%
% The matrix is given by
%    A(i,j) = delta(i,j) + h*beta*log|xi - xj|
% where "xi" is a grid point on the surface, and "h" is the mesh size.
%
% The matrix A sort of emulates a single layer potential, but note 
% that there are no quadratures involved.


function main51()

acc = 1e-6;
flag_pot = [];
params = [];

h     = 0.02;
r     = 0.3;

%n1 = round(  2*pi/h);
%n2 = round(r*2*pi/h);
n1 = 80;
n2 = 40;
fprintf(1,'n1   = %4d\n',n1);
fprintf(1,'n2   = %4d\n',n2);
fprintf(1,'acc  = %12.2e\n',acc);
[tt1,tt2] = meshgrid(linspace(0,2*pi,n1+1),linspace(0,r*2*pi,n2+1));
tt1       = reshape(tt1,1,numel(tt1));
tt2       = reshape(tt2,1,numel(tt2));
%tri       = delaunay(tt1,tt2);
tri       = zeros(5*n1*n2,3);
ntot = 0;
for i1 = 1:n1
  for i2 = 1:n2
    i_sw = (i1-1)*(n2+1) + i2;
    i_nw = (i1-1)*(n2+1) + i2 + 1;
    i_se = (i1-0)*(n2+1) + i2;
    i_ne = (i1-0)*(n2+1) + i2 + 1;
    tri(ntot + 1,:) = [i_sw,i_nw,i_se];
    tri(ntot + 2,:) = [i_nw,i_ne,i_se];
    ntot = ntot + 2;
  end
end
tri = tri(1:ntot,:);
fprintf(1,'ntot = %4d\n',ntot);

xx1 = cos(tt1).*(1 + r*cos(tt2/r));
xx2 = sin(tt1).*(1 + r*cos(tt2/r)).*(1 + 0.1*cos(tt1).*cos(tt2/r));
xx3 = 0.2*sin(tt2/r).*(1 + 0.1*cos(3*tt1));

xxc = [(1/3)*sum(xx1(tri'));...
       (1/3)*sum(xx2(tri'));...
       (1/3)*sum(xx3(tri'))];
ttc = [(1/3)*sum(tt1(tri'));...
       (1/3)*sum(tt2(tri'))];

box_geom = [0,2*pi,0,r*2*pi];
nmax = 100;

%%% Order the points in a tree.
[NODES,nlevels,indvec] = LOCAL_get_tree(ttc,box_geom,nmax);
tri = tri(indvec,:);
xxc = xxc(:,indvec);

%tic
%NODES  = LOCAL_compress_HSS_fsym_brute(xxc,NODES,acc,flag_pot,params);
%fprintf(1,'Time for full  compression = %f\n',toc)
tic
NODES = LOCAL_compress_HSS_fsym_green(xxc,NODES,acc,flag_pot,params);
fprintf(1,'Time for Green compression = %f\n',toc)

qq = randn(ntot,1);
tic
uu = OMNI_apply_HSS_fsym(NODES,qq);
fprintf(1,'Time for applying the matrix = %f\n',toc)
qq = randn(ntot,3);
uu = OMNI_apply_HSS_fsym(NODES,qq);
if (ntot < 4000)
  A        = LOCAL_construct_A_diag(xxc,1:ntot,flag_pot,params);
  uu_exact = A*qq;
  err      = norm(uu - uu_exact)/norm(uu_exact);
  fprintf(1,'Compression error = %12.3e\n',err)
else
  n_samp = 20;
  [trash,ind_samp] = sort(rand(1,ntot)) ;
  ind_samp = ind_samp(1:n_samp);
  uu_exact = LOCAL_applyrows(xxc,flag_pot,params,ind_samp,qq);
  fprintf(1,'Relative error (sampled error vector) = %13.6e\n',norm(uu_exact - uu(ind_samp,:))/norm(uu_exact))
end

figure(1)
hold off
%plot3(xx1,xx2,xx3,'.k')
trisurf(tri,xx1,xx2,xx3)
%plot3(xx1(tri(:,[1,2,3,1])'),xx2(tri(:,[1,2,3,1])'),xx3(tri(:,[1,2,3,1])'),'k')
axis equal
title('The domain in physical space')

figure(2)
hold off
plot(tt1,tt2,'k.',...
     tt1(tri(:,[1,2,3,1])'),tt2(tri(:,[1,2,3,1])'),'k')
axis equal
axis([-0.2,2*pi+0.2,-0.2,2.1])
title('The domain in parameter space')
keyboard

INDS     = cell(1,nlevels);
LEVELIND = cell(1,nlevels);
for ibox = 2:size(NODES,2)
  ilevel = NODES{2,ibox};
  INDS{ilevel} = [INDS{ilevel},NODES{20,ibox}];
  LEVELIND{ilevel} = [LEVELIND{ilevel},ibox];
end

for ilevel = min(nlevels,6):(-1):1
  fprintf(1,'ilevel = %2d     average rank = %6.1f    points = %5d\n',...
          ilevel,length(INDS{ilevel})/length(LEVELIND{ilevel}),length(INDS{ilevel}));
  figure(1)
  hold off
  trisurf(tri(INDS{ilevel},:),xx1,xx2,xx3)
  axis equal
  title('The domain in physical space')
  figure(2)
  hold off
  plot(tt1(tri(INDS{ilevel},[1,2,3,1])'),...
       tt2(tri(INDS{ilevel},[1,2,3,1])'),'k')
  axis equal
  axis([-0.2,2*pi+0.2,-0.2,2.1])
  hold on
  for ibox = LEVELIND{ilevel}
    plot(NODES{1,ibox}([1,2,2,1,1]),NODES{1,ibox}([3,3,4,4,3]),'r')
  end
  title('The domain in parameter space')
  keyboard
end


return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [NODES,nlevels,indvec] = LOCAL_get_tree(xx,box_geom,nmax)

ntot = size(xx,2);

% BOXES is a temporary work array for setting up the tree-structure
BOXES       = zeros(18,100);
lenBOXES    = 100;
BOXES( 1,1) = NaN;
BOXES( 2,1) = 0;
BOXES( 3,1) = NaN;
BOXES( 6,1) = 1;
BOXES( 7,1) = ntot;
BOXES(10,1) = NaN;
BOXES(11,1) = box_geom(1);
BOXES(12,1) = box_geom(2);
BOXES(13,1) = box_geom(3);
BOXES(14,1) = box_geom(4);
BOXES(15,1) = -1;
BOXES(16,1) = -1;
BOXES(17,1) = -1;
BOXES(18,1) = -1;

indvec = 1:ntot;

% Create the tree structure by quartering any
% box that holds more than nmax nodes.
%
% We create the boxes one level at a time.
% The following loop is over LEVELS.
ibox_last  = 0;
ibox_new   = 1;
ilevel     = 0;
while (ibox_new > ibox_last)

  ibox_first = ibox_last+1;
  ibox_last  = ibox_new;
  ilevel     = ilevel+1;
  
  % Loop over all boxes on the level that was last created.
  % All newly created boxes are temporarily stored in the array TMPBOXES.
  for ibox = ibox_first:ibox_last
    
    % If ibox holds more than nmax nodes, it will be partitioned.
    if (BOXES(07,ibox) > nmax) 

      qfirst = BOXES(06,ibox);
      x1min  = BOXES(11,ibox);
      x1max  = BOXES(12,ibox);
      x2min  = BOXES(13,ibox);
      x2max  = BOXES(14,ibox);
      J      = qfirst - 1 + (1:BOXES(07,ibox));
      indloc = indvec(J);
      if ((x1max - x1min) >= (x2max - x2min)) % Make a vertical cut
        x1half = (x1min + x1max)/2;
        J_son1 = find(xx(1,indloc) <= x1half);
        J_son2 = find(xx(1,indloc) >  x1half);
        box_geom_son1 = [x1min,x1half,x2min,x2max];
        box_geom_son2 = [x1half,x1max,x2min,x2max];
      else                                   % Make a horizontal cut
        x2half = (x2min + x2max)/2;
        J_son1 = find(xx(2,indloc) <= x2half);
        J_son2 = find(xx(2,indloc) >  x2half);
        box_geom_son1 = [x1min,x1max,x2min,x2half];
        box_geom_son2 = [x1min,x1max,x2half,x2max];
      end
%      plot(box_geom([1,2,2,1,1]),box_geom([3,3,4,4,3]),'k',...
%           box_geom_son1([1,2,2,1,1]),box_geom_son1([3,3,4,4,3]),'r',...
%           box_geom_son2([1,2,2,1,1]),box_geom_son2([3,3,4,4,3]),'b',...
%           xx(1,indloc(J_son1)),xx(2,indloc(J_son1)),'r.',...
%           xx(1,indloc(J_son2)),xx(2,indloc(J_son2)),'bo')
%      legend('big box','son1','son2','son1 points','son2 points')
%      axis equal
%      keyboard
      
      indvec(J) = [indloc(J_son1),indloc(J_son2)];
      
      npart_added = 0;
      % If there is not enough space to save the 2 children in
      % the array BOXES, then double the size of BOXES.
      if ((ibox_new + 2) > lenBOXES)
        BOXES  = [BOXES,zeros(size(BOXES,1),4+size(BOXES,2))];
        lenBOXES = size(BOXES,2);
      end
      
      if ~isempty(J_son1)
        ibox_new = ibox_new + 1;
        n_son1   = length(J_son1);
        BOXES(:,ibox_new) = [NaN,ilevel,ibox,NaN,NaN,qfirst+npart_added,...
                             n_son1,NaN,NaN,NaN,box_geom_son1,...
                             -1,-1,-1,-1]';
        BOXES(15,ibox) = ibox_new;
        npart_added = npart_added + n_son1;
      end
      if ~isempty(J_son2)
        ibox_new = ibox_new + 1;
        n_son2 = length(J_son2);
        BOXES(:,ibox_new) = [NaN,ilevel,ibox,NaN,NaN,qfirst+npart_added,...
                             n_son2,NaN,NaN,NaN,box_geom_son2,...
                             -1,-1,-1,-1]';
        BOXES(16,ibox) = ibox_new;
        npart_added = npart_added + n_son2;
      end
    
    end
    
  end
  
end

nlevels = ilevel - 1;

% Let nboxes denote the number of boxes created,
% create the object NODES which will hold all
% relevant information, and transfer the 
% information in BOXES to NODES.
% We also delete the object BOXES.
nboxes  = ibox_new;
NODES = cell(45,nboxes);
for ibox = 1:nboxes
  NODES{01,ibox} = BOXES(11:14,ibox);
  NODES{02,ibox} = BOXES(02,ibox);
  NODES{03,ibox} = BOXES(03,ibox);
  NODES{04,ibox} = [];
  for j = 15:16
    if (BOXES(j,ibox) > 0)
      NODES{04,ibox} = [NODES{04,ibox},BOXES(j,ibox)];
    end
  end
  NODES{05,ibox} = length(NODES{04,ibox});
  NODES{06,ibox} = BOXES(06,ibox);
  NODES{07,ibox} = BOXES(07,ibox);
end


return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_compress_HSS_fsym_brute(xxc,NODES,acc,flag_pot,params)

% Extract parameters.
ntot   = size(xxc,2);
nboxes = size(NODES,2);

% Compress all nodes, going from smaller to larger.
for ibox = nboxes:(-1):2
  ind_offd = [1:(NODES{6,ibox}-1),(NODES{6,ibox}+NODES{7,ibox}):ntot];
  if (NODES{5,ibox}==0) % ibox has no sons.
    ind_skel = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
  elseif (NODES{5,ibox}==1) % ibox has a son
    ison     = NODES{4,ibox};
    ind_skel = NODES{20,ison};
  else % ibox has two sons
    ison1    = NODES{4,ibox}(1);
    ison2    = NODES{4,ibox}(2);
    ind_skel = [NODES{20,ison1},NODES{20,ison2}];
  end
  A21 = LOCAL_construct_A_offd(xxc,ind_offd,ind_skel,flag_pot,params);
  % Compute the skeletons.
  [T,J] = id_decomp(A21,acc);
  k = size(T,1);
  % Record the outgoing skeletons:
  NODES{20,ibox} = ind_skel(J(1:k));
  NODES{30,ibox} = T;
  NODES{31,ibox} = J;
  % Record nskel and kskel
  NODES{ 8,ibox} = length(ind_skel);
  NODES{ 9,ibox} = k;
  fprintf(1,'ibox = %3d    n = %4d    k = %3d\n',ibox,length(ind_skel),k)
%  figure(1)
%  hold off
%  JJ = ind_skel(J(1:k));
%  plot3(xxc(1,:),xxc(2,:),xxc(3,:),'k.',...
%        xxc(1,JJ),xxc(2,JJ),xxc(3,JJ),'r.')
%  axis equal
%  keyboard
end

% Construct the matrices representing self-interactions on the leaves.
for ibox = nboxes:(-1):2
  if (NODES{5,ibox}==0)
    ind = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
    NODES{40,ibox} = LOCAL_construct_A_diag(xxc,ind,flag_pot,params);
  end
end

% Construct the matrices for sibling interactions.
for ibox = nboxes:(-1):1
  if (NODES{5,ibox}==2)
    ison1 = NODES{4,ibox}(1);
    ison2 = NODES{4,ibox}(2);
    NODES{46,ison1} = LOCAL_construct_A_offd(xxc,NODES{20,ison1},NODES{20,ison2},flag_pot,params);
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_compress_HSS_fsym_green(xxc,NODES,acc,flag_pot,params)

% Construct the proxy surface.
r_sep = 1.5; % Separation parameter in the "Green trick".
             % It specifies how much larger the proxy sphere is.
xxproxy_ref = LOCAL_get_xxproxy_ref(r_sep,acc);
fprintf(1,'nproxy = %d\n',size(xxproxy_ref,2))

% Extract parameters.
nboxes = size(NODES,2);

NODES    = LOCAL_find_spheres(xxc,NODES);
LIST_NEI = LOCAL_construct_potential_neighborlist(xxc,NODES,r_sep);

% Compress all nodes, going from smaller to larger.
for ibox = nboxes:(-1):2
%  ind_offd = [1:(NODES{6,ibox}-1),(NODES{6,ibox}+NODES{7,ibox}):ntot];
  if (NODES{5,ibox}==0) % ibox has no sons.
    ind_skel = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
  elseif (NODES{5,ibox}==1) % ibox has a son
    ison     = NODES{4,ibox};
    ind_skel = NODES{20,ison};
  else % ibox has two sons
    ison1    = NODES{4,ibox}(1);
    ison2    = NODES{4,ibox}(2);
    ind_skel = [NODES{20,ison1},NODES{20,ison2}];
  end
  ind_near = LOCAL_get_ind_near(xxc,NODES,ibox,LIST_NEI,r_sep);
  cc       = NODES{1,ibox}(5:7);
  R        = NODES{1,ibox}(8);
  xxproxy  = cc*ones(1,size(xxproxy_ref,2)) + R*xxproxy_ref;
  B21 = LOCAL_construct_A_proxy(xxproxy,xxc,ind_skel,flag_pot,params);
  C21 = LOCAL_construct_A_offd(xxc,ind_near,ind_skel,flag_pot,params);
%  A21 = LOCAL_construct_A_offd(xxc,ind_offd,ind_skel,flag_pot,params);
  % Compute the skeletons.
  [T,J] = id_decomp([B21;C21],acc);
%  [T,J] = id_decomp(A21,acc);
  k = size(T,1);
  % Record the outgoing skeletons:
  NODES{20,ibox} = ind_skel(J(1:k));
  NODES{30,ibox} = T;
  NODES{31,ibox} = J;
  % Record nskel and kskel
  NODES{ 8,ibox} = length(ind_skel);
  NODES{ 9,ibox} = k;
  fprintf(1,'ibox = %3d    n = %4d    k = %3d\n',ibox,length(ind_skel),k)
%  figure(1)
%  ntot = size(xxc,2);
%  ind_offd = [1:(NODES{6,ibox}-1),(NODES{6,ibox}+NODES{7,ibox}):ntot];
%  plot3(xxc(1,ind_offd),xxc(2,ind_offd),xxc(3,ind_offd),'k.',...
%        xxc(1,ind_skel),xxc(2,ind_skel),xxc(3,ind_skel),'b.',...
%        xxc(1,ind_near),xxc(2,ind_near),xxc(3,ind_near),'r.',...
%        xxproxy(1,:),xxproxy(2,:),xxproxy(3,:),'g.')
%  legend('ind_offd','ind_skel','ind_near','proxy')
%  axis equal
%  keyboard
end

% Construct the matrices representing self-interactions on the leaves.
for ibox = nboxes:(-1):2
  if (NODES{5,ibox}==0)
    ind = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
    NODES{40,ibox} = LOCAL_construct_A_diag(xxc,ind,flag_pot,params);
  end
end

% Construct the matrices for sibling interactions.
for ibox = nboxes:(-1):1
  if (NODES{5,ibox}==2)
    ison1 = NODES{4,ibox}(1);
    ison2 = NODES{4,ibox}(2);
    NODES{46,ison1} = LOCAL_construct_A_offd(xxc,NODES{20,ison1},NODES{20,ison2},flag_pot,params);
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_construct_A_diag(xx,ind,flag_pot,params)

ntot = size(xx,2);
nloc = length(ind);

dsq = (xx(1,ind)' * ones(1,nloc) - ones(nloc,1) * xx(1,ind)).^2 + ...
      (xx(2,ind)' * ones(1,nloc) - ones(nloc,1) * xx(2,ind)).^2 + ...
      (xx(3,ind)' * ones(1,nloc) - ones(nloc,1) * xx(3,ind)).^2 + ...
      eye(nloc);
  
A   = (0.5/ntot)*log(dsq);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A12 = LOCAL_construct_A_offd(xx,ind1,ind2,flag_pot,params)

ntot = size(xx,2);
n1   = length(ind1);
n2   = length(ind2);
dsq  = (xx(1,ind1)' * ones(1,n2) - ones(n1,1) * xx(1,ind2)).^2 + ...
       (xx(2,ind1)' * ones(1,n2) - ones(n1,1) * xx(2,ind2)).^2 + ...
       (xx(3,ind1)' * ones(1,n2) - ones(n1,1) * xx(3,ind2)).^2;

A12  = (0.5/ntot)*log(dsq);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function B = LOCAL_construct_A_proxy(xxproxy,xx,ind2,flag_pot,params)

ntot = size(xx,2);
n1   = size(xxproxy,2);
n2   = length(ind2);
dsq  = (xxproxy(1,:)' * ones(1,n2) - ones(n1,1) * xx(1,ind2)).^2 + ...
       (xxproxy(2,:)' * ones(1,n2) - ones(n1,1) * xx(2,ind2)).^2 + ...
       (xxproxy(3,:)' * ones(1,n2) - ones(n1,1) * xx(3,ind2)).^2;
B    = (0.5/ntot)*log(dsq);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function uu = LOCAL_applyrows(xx,flag_pot,params,ind_samp,qq)

if (length(xx)*length(ind_samp) > 9000000)
  disp('in OMNI_applyrows')
  disp('The problem size seems a bit large ... breaking.')
  keyboard
end

ntot   = length(xx);
indtmp = 1:ntot;
indtmp(ind_samp) = 2*ntot;
indtmp = sort(indtmp);
ind_offd = indtmp(1:(ntot-length(ind_samp)));

A_offd = LOCAL_construct_A_offd(xx,ind_samp,ind_offd,flag_pot,params);
A_diag = LOCAL_construct_A_diag(xx,ind_samp,flag_pot,params);

uu = A_offd*qq(ind_offd,:) + A_diag*qq(ind_samp,:);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LIST_NEI = LOCAL_construct_potential_neighborlist(xxc,NODES,r_sep)

nboxes = size(NODES,2);

% Note that in the construction, we temporarily include
% a box in its own list of neighbors.
LIST_NEI    = cell(1,nboxes);
LIST_NEI{1} = [1];
for ibox = 2:nboxes
  box_geom = NODES{1,ibox};
  cc       = box_geom(5:7);
  R        = box_geom(8);
  % Collect of list of all the father's neighbors' sons:
  potneis = [];
  for jbox = LIST_NEI{NODES{3,ibox}};
    potneis = [potneis,NODES{4,jbox}];
  end
  % Loop over all potential neighbors to check which ones are actually "close".
  for jbox = potneis
    ind      = NODES{6,jbox} - 1 + (1:NODES{7,jbox});
    distsq   = (xxc(1,ind) - cc(1)).*(xxc(1,ind) - cc(1)) + ...
               (xxc(2,ind) - cc(2)).*(xxc(2,ind) - cc(2)) + ...
               (xxc(3,ind) - cc(3)).*(xxc(3,ind) - cc(3));
    J_inside = (distsq < (r_sep*R)^2);
    if (sum(J_inside) > 0)
      LIST_NEI{ibox} = [LIST_NEI{ibox},jbox];
    end
  end         
end
% Finally we remove the box itself from the list of neighbors.
for ibox = 2:nboxes
  j = find(LIST_NEI{ibox} == ibox);
  LIST_NEI{ibox} = sort(LIST_NEI{ibox}([1:(j-1),(j+1):end]));
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ind_near = LOCAL_get_ind_near(xxc,NODES,ibox,LIST_NEI,r_sep)

box_geom = NODES{1,ibox};
cc       = box_geom(5:7);
R        = box_geom(8);
ind_near = [];
for jbox = LIST_NEI{ibox}
  if (NODES{5,jbox} == 0)
    indj     = NODES{6,jbox} - 1 + (1:NODES{7,jbox});
  else
    indj     = [];
    for kbox = NODES{4,jbox}
      indj = [indj,NODES{20,kbox}];
    end
  end
  distsq   = (xxc(1,indj) - cc(1)).*(xxc(1,indj) - cc(1)) + ...
             (xxc(2,indj) - cc(2)).*(xxc(2,indj) - cc(2)) + ...
             (xxc(3,indj) - cc(3)).*(xxc(3,indj) - cc(3));
  J_inside = (distsq < (r_sep*R)^2);
  ind_near = [ind_near,indj(J_inside)];
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_find_spheres(xxc,NODES)

for ibox = 1:size(NODES,2)
  ind   = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
  xxloc = xxc(:,ind);
  nbox  = length(ind);
  cc    = (1/nbox)*(sum(xxloc')');
  rsq   = (xxloc(1,:) - cc(1)).*(xxloc(1,:) - cc(1)) + ...
          (xxloc(2,:) - cc(2)).*(xxloc(2,:) - cc(2)) + ...
          (xxloc(3,:) - cc(3)).*(xxloc(3,:) - cc(3));
  R     = sqrt(max(rsq));
  NODES{1,ibox} = [NODES{1,ibox};cc;R];
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ss1,ss2,ss3] = LOCAL_get_sphere(cc,R,n1,n2)

if (nargin == 2)
  n1 = 20;
  n2 = 40;
end
  
[tt1,tt2] = meshgrid(linspace(0,pi,40),linspace(0,2*pi,40));
ss1       = cc(1) + R*sin(tt1).*cos(tt2); 
ss2       = cc(2) + R*sin(tt1).*sin(tt2); 
ss3       = cc(3) + R*cos(tt1);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function xxproxy = LOCAL_get_xxproxy_ref(r_sep,acc)

n_init = 1000;

xxproxy     = randn(3,n_init);
xxproxy_in  = xxproxy.*(ones(3,1)*(1./sqrt(sum(xxproxy.*xxproxy))));
xxproxy_out = r_sep*xxproxy_in;

dist1 = xxproxy_in(1,:)' * ones(1,n_init) - ones(n_init,1) * xxproxy_out(1,:);
dist2 = xxproxy_in(2,:)' * ones(1,n_init) - ones(n_init,1) * xxproxy_out(2,:);
dist3 = xxproxy_in(3,:)' * ones(1,n_init) - ones(n_init,1) * xxproxy_out(3,:);
        
distsq = dist1.*dist1 + dist2.*dist2 + dist3.*dist3;

B = (0.5/n_init)*log(distsq);
[T,J] = id_decomp(B,0.2*acc);
k     = size(T,1);
xxproxy = xxproxy_out(:,J(1:k));

return
